@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2021 gnome-mpv
+ * Copyright (c) 2014-2022 gnome-mpv
  *
  * This file is part of Celluloid.
  *
@@ -30,6 +30,8 @@ enum
 	PROP_SKIP_ENABLED,
 	PROP_DURATION,
 	PROP_ENABLED,
+	PROP_COMPACT,
+	PROP_FULLSCREENED,
 	PROP_PAUSE,
 	PROP_SHOW_FULLSCREEN_BUTTON,
 	PROP_TIME_POSITION,
@@ -38,12 +40,17 @@ enum
 	PROP_VOLUME,
 	PROP_VOLUME_MAX,
 	PROP_VOLUME_POPUP_VISIBLE,
+	PROP_CHAPTER_LIST,
 	N_PROPERTIES
 };
 
 struct _CelluloidControlBox
 {
 	GtkBox parent_instance;
+
+	GtkGesture *click_gesture;
+
+	GtkWidget *inner_box;
 	GtkWidget *play_button;
 	GtkWidget *stop_button;
 	GtkWidget *forward_button;
@@ -55,9 +62,12 @@ struct _CelluloidControlBox
 	GtkWidget *volume_button;
 	GtkWidget *fullscreen_button;
 	GtkWidget *seek_bar;
+	GtkWidget *secondary_seek_bar;
 	gboolean skip_enabled;
 	gdouble duration;
 	gboolean enabled;
+	gboolean compact;
+	gboolean fullscreened;
 	gboolean pause;
 	gboolean show_fullscreen_button;
 	gdouble time_position;
@@ -66,6 +76,7 @@ struct _CelluloidControlBox
 	gdouble volume;
 	gdouble volume_max;
 	gboolean volume_popup_visible;
+	GPtrArray *chapter_list;
 };
 
 struct _CelluloidControlBoxClass
@@ -85,6 +96,9 @@ get_property(	GObject *object,
 		GValue *value,
 		GParamSpec *pspec );
 
+static void
+css_changed(GtkWidget *self, GtkCssStyleChange *change);
+
 static gboolean
 gtk_to_mpv_volume(	GBinding *binding,
 			const GValue *from_value,
@@ -103,6 +117,13 @@ seek_handler(CelluloidSeekBar *seek_bar, gdouble value, gpointer data);
 static void
 volume_changed_handler(GtkVolumeButton *button, gdouble value, gpointer data);
 
+static gboolean
+button_pressed_handler(	GtkEventControllerKey *controller,
+			guint keyval,
+			guint keycode,
+			GdkModifierType state,
+			gpointer data );
+
 static void
 simple_signal_handler(GtkWidget *widget, gpointer data);
 
@@ -115,6 +136,12 @@ set_skip_enabled(CelluloidControlBox *box, gboolean enabled);
 static void
 set_playing_state(CelluloidControlBox *box, gboolean playing);
 
+static void
+set_compact(CelluloidControlBox *box, gboolean compact);
+
+static void
+set_fullscreen_state(CelluloidControlBox *box, gboolean fullscreen);
+
 static void
 init_button(	GtkWidget *button,
 		const gchar *icon_name,
@@ -139,9 +166,6 @@ set_property(	GObject *object,
 
 		case PROP_DURATION:
 		self->duration = g_value_get_double(value);
-
-		celluloid_seek_bar_set_duration
-			(CELLULOID_SEEK_BAR(self->seek_bar), self->duration);
 		break;
 
 		case PROP_ENABLED:
@@ -149,6 +173,16 @@ set_property(	GObject *object,
 		set_enabled(self, self->enabled);
 		break;
 
+		case PROP_COMPACT:
+		self->compact = g_value_get_boolean(value);
+		set_compact(self, self->compact);
+		break;
+
+		case PROP_FULLSCREENED:
+		self->fullscreened = g_value_get_boolean(value);
+		set_fullscreen_state(self, self->fullscreened);
+		break;
+
 		case PROP_PAUSE:
 		self->pause = g_value_get_boolean(value);
 		set_playing_state(self, !self->pause);
@@ -156,6 +190,9 @@ set_property(	GObject *object,
 
 		case PROP_SHOW_FULLSCREEN_BUTTON:
 		self->show_fullscreen_button = g_value_get_boolean(value);
+		gtk_widget_set_visible
+			(	self->fullscreen_button,
+				self->show_fullscreen_button );
 		break;
 
 		case PROP_TIME_POSITION:
@@ -164,6 +201,9 @@ set_property(	GObject *object,
 		celluloid_seek_bar_set_pos
 			(	CELLULOID_SEEK_BAR(self->seek_bar),
 				self->time_position );
+		celluloid_seek_bar_set_pos
+			(	CELLULOID_SEEK_BAR(self->secondary_seek_bar),
+				self->time_position );
 		break;
 
 		case PROP_LOOP:
@@ -196,6 +236,10 @@ set_property(	GObject *object,
 		}
 		break;
 
+		case PROP_CHAPTER_LIST:
+		self->chapter_list = g_value_get_pointer(value);
+		break;
+
 		default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
 		break;
@@ -224,6 +268,14 @@ get_property(	GObject *object,
 		g_value_set_boolean(value, self->enabled);
 		break;
 
+		case PROP_COMPACT:
+		g_value_set_boolean(value, self->compact);
+		break;
+
+		case PROP_FULLSCREENED:
+		g_value_set_boolean(value, self->fullscreened);
+		break;
+
 		case PROP_PAUSE:
 		g_value_set_boolean(value, self->pause);
 		break;
@@ -256,12 +308,31 @@ get_property(	GObject *object,
 		g_value_set_boolean(value, self->volume_popup_visible);
 		break;
 
+		case PROP_CHAPTER_LIST:
+		g_value_set_pointer(value, self->chapter_list);
+		break;
+
 		default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
 		break;
 	}
 }
 
+static void
+css_changed(GtkWidget *self, GtkCssStyleChange *change)
+{
+	GTK_WIDGET_CLASS(celluloid_control_box_parent_class)
+		->css_changed(self, change);
+
+	CelluloidControlBox *box = CELLULOID_CONTROL_BOX(self);
+	GtkStyleContext *ctx = gtk_widget_get_style_context(self);
+	GtkBorder padding = {0};
+
+	gtk_style_context_get_padding(ctx, &padding);
+	g_object_set(box->seek_bar, "popover-y-offset", -padding.top, NULL);
+	g_object_set(box->secondary_seek_bar, "popover-y-offset", -padding.top, NULL);
+}
+
 static gboolean
 gtk_to_mpv_volume(	GBinding *binding,
 			const GValue *from_value,
@@ -298,6 +369,16 @@ volume_changed_handler(GtkVolumeButton *button, gdouble value, gpointer data)
 	g_signal_emit_by_name(data, "volume-changed", value*100.0);
 }
 
+static gboolean
+button_pressed_handler(	GtkEventControllerKey *controller,
+			guint keyval,
+			guint keycode,
+			GdkModifierType state,
+			gpointer data )
+{
+	return TRUE;
+}
+
 static void
 simple_signal_handler(GtkWidget *widget, gpointer data)
 {
@@ -345,8 +426,8 @@ set_skip_enabled(CelluloidControlBox *box, gboolean enabled)
 {
 	if(enabled)
 	{
-		gtk_widget_show_all(box->previous_button);
-		gtk_widget_show_all(box->next_button);
+		gtk_widget_show(box->previous_button);
+		gtk_widget_show(box->next_button);
 	}
 	else
 	{
@@ -358,16 +439,45 @@ set_skip_enabled(CelluloidControlBox *box, gboolean enabled)
 static void
 set_playing_state(CelluloidControlBox *box, gboolean playing)
 {
-	GtkWidget *image = gtk_button_get_image(GTK_BUTTON(box->play_button));
-	const gchar *tooltip = playing?_("Pause"):_("Play");
-
-	gtk_image_set_from_icon_name( 	GTK_IMAGE(image),
-					playing?
-					"media-playback-pause-symbolic":
-					"media-playback-start-symbolic",
-					GTK_ICON_SIZE_BUTTON );
+	const gchar *tooltip = playing ? _("Pause") : _("Play");
+	const gchar *icon_name =
+		playing ?
+		"media-playback-pause-symbolic" :
+		"media-playback-start-symbolic";
 
 	gtk_widget_set_tooltip_text(box->play_button, tooltip);
+	gtk_button_set_icon_name(GTK_BUTTON(box->play_button), icon_name);
+}
+
+static void
+set_compact(CelluloidControlBox *box, gboolean compact)
+{
+	if(box->compact)
+	{
+		gtk_widget_set_halign(box->inner_box, GTK_ALIGN_CENTER);
+		gtk_widget_set_margin_start(box->loop_button, 12);
+		gtk_widget_hide(box->seek_bar);
+		gtk_widget_show(box->secondary_seek_bar);
+	}
+	else
+	{
+		gtk_widget_set_halign(box->inner_box, GTK_ALIGN_FILL);
+		gtk_widget_set_margin_start(box->loop_button, 0);
+		gtk_widget_show(box->seek_bar);
+		gtk_widget_hide(box->secondary_seek_bar);
+	}
+}
+
+static void
+set_fullscreen_state(CelluloidControlBox *box, gboolean fullscreen)
+{
+	const gchar *icon_name =
+		fullscreen ?
+		"view-restore-symbolic" :
+		"view-fullscreen-symbolic";
+
+	gtk_button_set_icon_name
+		(GTK_BUTTON(box->fullscreen_button), icon_name);
 }
 
 static void
@@ -375,23 +485,23 @@ init_button(	GtkWidget *button,
 		const gchar *icon_name,
 		const gchar *tooltip_text )
 {
-	GtkWidget *icon =	gtk_image_new_from_icon_name
-				(icon_name, GTK_ICON_SIZE_BUTTON);
+	gtk_button_set_icon_name(GTK_BUTTON(button), icon_name);
 
+	g_object_set(button, "has-frame", FALSE, NULL);
 	gtk_widget_set_tooltip_text(button, tooltip_text);
-	g_object_set(button, "relief", GTK_RELIEF_NONE, NULL);
 	gtk_widget_set_can_focus(button, FALSE);
-	gtk_button_set_image(GTK_BUTTON(button), icon);
 }
 
 static void
 celluloid_control_box_class_init(CelluloidControlBoxClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS(klass);
+	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
 	GParamSpec *pspec = NULL;
 
 	object_class->set_property = set_property;
 	object_class->get_property = get_property;
+	widget_class->css_changed = css_changed;
 
 	pspec = g_param_spec_boolean
 		(	"skip-enabled",
@@ -422,6 +532,22 @@ celluloid_control_box_class_init(CelluloidControlBoxClass *klass)
 	g_object_class_install_property
 		(object_class, PROP_ENABLED, pspec);
 
+	pspec = g_param_spec_boolean
+		(	"compact",
+			"Compact",
+			"Whether or not compact mode is enabled",
+			FALSE,
+			G_PARAM_READWRITE );
+	g_object_class_install_property(object_class, PROP_COMPACT, pspec);
+
+	pspec = g_param_spec_boolean
+		(	"fullscreened",
+			"Fullscreened",
+			"Whether the control box is in fullscreen configuration",
+			FALSE,
+			G_PARAM_READWRITE );
+	g_object_class_install_property(object_class, PROP_FULLSCREENED, pspec);
+
 	pspec = g_param_spec_boolean
 		(	"pause",
 			"Pause",
@@ -491,6 +617,14 @@ celluloid_control_box_class_init(CelluloidControlBoxClass *klass)
 	g_object_class_install_property
 		(object_class, PROP_VOLUME_MAX, pspec);
 
+	pspec = g_param_spec_pointer
+		(	"chapter-list",
+			"Chapter list",
+			"The list of chapters for the current file",
+			G_PARAM_READWRITE );
+	g_object_class_install_property
+		(object_class, PROP_CHAPTER_LIST, pspec);
+
 	pspec = g_param_spec_boolean
 		(	"volume-popup-visible",
 			"Volume popup visible",
@@ -538,6 +672,9 @@ celluloid_control_box_init(CelluloidControlBox *box)
 	GtkWidget *popup = NULL;
 	GtkAdjustment *adjustment = NULL;
 
+	box->click_gesture = gtk_gesture_click_new();
+
+	box->inner_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
 	box->play_button = gtk_button_new();
 	box->stop_button = gtk_button_new();
 	box->forward_button = gtk_button_new();
@@ -549,9 +686,12 @@ celluloid_control_box_init(CelluloidControlBox *box)
 	box->shuffle_button = gtk_toggle_button_new();
 	box->volume_button = gtk_volume_button_new();
 	box->seek_bar = celluloid_seek_bar_new();
+	box->secondary_seek_bar = celluloid_seek_bar_new();
 	box->skip_enabled = FALSE;
 	box->duration = 0.0;
 	box->enabled = TRUE;
+	box->compact = FALSE;
+	box->fullscreened = FALSE;
 	box->pause = TRUE;
 	box->show_fullscreen_button = FALSE;
 	box->time_position = 0.0;
@@ -559,6 +699,8 @@ celluloid_control_box_init(CelluloidControlBox *box)
 	box->shuffle = FALSE;
 	box->volume = 0.0;
 	box->volume_max = 100.0;
+	box->volume_popup_visible = FALSE;
+	box->chapter_list = NULL;
 
 	init_button(	box->play_button,
 			"media-playback-start-symbolic",
@@ -588,35 +730,90 @@ celluloid_control_box_init(CelluloidControlBox *box)
 			"view-fullscreen-symbolic",
 			_("Toggle Fullscreen") );
 
+	gtk_widget_add_controller
+		(GTK_WIDGET(box), GTK_EVENT_CONTROLLER(box->click_gesture));
+
 	gtk_style_context_add_class
-		(	gtk_widget_get_style_context(GTK_WIDGET(box)),
-			GTK_STYLE_CLASS_BACKGROUND );
+		(gtk_widget_get_style_context(GTK_WIDGET(box)), "toolbar");
 
 	gtk_widget_set_margin_end(box->seek_bar, 6);
+	gtk_widget_set_margin_end(box->secondary_seek_bar, 6);
 
 	gtk_widget_set_can_focus(box->volume_button, FALSE);
 	gtk_widget_set_can_focus(box->seek_bar, FALSE);
+	gtk_widget_set_can_focus(box->secondary_seek_bar, FALSE);
 
-	gtk_container_add(GTK_CONTAINER(box), box->previous_button);
-	gtk_container_add(GTK_CONTAINER(box), box->rewind_button);
-	gtk_container_add(GTK_CONTAINER(box), box->play_button);
-	gtk_container_add(GTK_CONTAINER(box), box->stop_button);
-	gtk_container_add(GTK_CONTAINER(box), box->forward_button);
-	gtk_container_add(GTK_CONTAINER(box), box->next_button);
-	gtk_box_pack_start(GTK_BOX(box), box->seek_bar, TRUE, TRUE, 0);
-	gtk_container_add(GTK_CONTAINER(box), box->loop_button);
-	gtk_container_add(GTK_CONTAINER(box), box->shuffle_button);
-	gtk_container_add(GTK_CONTAINER(box), box->volume_button);
-	gtk_container_add(GTK_CONTAINER(box), box->fullscreen_button);
+	gtk_widget_set_hexpand(box->seek_bar, TRUE);
+	gtk_widget_set_hexpand(box->secondary_seek_bar, TRUE);
+
+	g_object_set(box->secondary_seek_bar, "show-label", FALSE, NULL);
+
+	set_compact(box, box->compact);
+
+	g_object_set(box, "orientation", GTK_ORIENTATION_VERTICAL, NULL);
+
+	gchar css_data[] = "box {border-spacing: 0px;}";
+	GtkCssProvider *css = gtk_css_provider_new();
+	gtk_css_provider_load_from_data(css, css_data, -1);
+
+	GtkStyleContext *style = gtk_widget_get_style_context(GTK_WIDGET(box));
+
+	gtk_style_context_add_provider
+		(	style,
+			GTK_STYLE_PROVIDER(css),
+			GTK_STYLE_PROVIDER_PRIORITY_USER );
+
+	box->inner_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
+
+	gtk_box_append(GTK_BOX(box), box->secondary_seek_bar);
+	gtk_box_append(GTK_BOX(box), box->inner_box);
+
+	gtk_box_append(GTK_BOX(box->inner_box), box->previous_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->rewind_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->play_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->stop_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->forward_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->next_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->seek_bar);
+	gtk_box_append(GTK_BOX(box->inner_box), box->loop_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->shuffle_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->volume_button);
+	gtk_box_append(GTK_BOX(box->inner_box), box->fullscreen_button);
 
 	popup =	gtk_scale_button_get_popup
 		(GTK_SCALE_BUTTON(box->volume_button));
 	adjustment =	gtk_scale_button_get_adjustment
 			(GTK_SCALE_BUTTON(box->volume_button));
 
+	gtk_popover_set_position(GTK_POPOVER(popup), GTK_POS_TOP);
+
 	g_object_bind_property(	popup, "visible",
 				box, "volume-popup-visible",
 				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "duration",
+				box->seek_bar, "duration",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "chapter-list",
+				box->seek_bar, "chapter-list",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "pause",
+				box->seek_bar, "pause",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "enabled",
+				box->seek_bar, "enabled",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "chapter-list",
+				box->secondary_seek_bar, "chapter-list",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "duration",
+				box->secondary_seek_bar, "duration",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "pause",
+				box->secondary_seek_bar, "pause",
+				G_BINDING_DEFAULT );
+	g_object_bind_property(	box, "enabled",
+				box->secondary_seek_bar, "enabled",
+				G_BINDING_DEFAULT );
 	g_object_bind_property(	box, "loop",
 				box->loop_button, "active",
 				G_BINDING_BIDIRECTIONAL );
@@ -638,10 +835,11 @@ celluloid_control_box_init(CelluloidControlBox *box)
 					NULL,
 					NULL );
 
-	g_signal_connect(	box,
-				"button-press-event",
-				G_CALLBACK(gtk_true),
+	g_signal_connect(	box->click_gesture,
+				"pressed",
+				G_CALLBACK(button_pressed_handler),
 				NULL );
+
 	g_signal_connect(	box->play_button,
 				"clicked",
 				G_CALLBACK(simple_signal_handler),
@@ -654,6 +852,10 @@ celluloid_control_box_init(CelluloidControlBox *box)
 				"seek",
 				G_CALLBACK(seek_handler),
 				box );
+	g_signal_connect(	box->secondary_seek_bar,
+				"seek",
+				G_CALLBACK(seek_handler),
+				box );
 	g_signal_connect(	box->forward_button,
 				"clicked",
 				G_CALLBACK(simple_signal_handler),
@@ -690,6 +892,7 @@ void
 celluloid_control_box_set_seek_bar_pos(CelluloidControlBox *box, gdouble pos)
 {
 	celluloid_seek_bar_set_pos(CELLULOID_SEEK_BAR(box->seek_bar), pos);
+	celluloid_seek_bar_set_pos(CELLULOID_SEEK_BAR(box->secondary_seek_bar), pos);
 }
 
 void
@@ -697,6 +900,7 @@ celluloid_control_box_set_seek_bar_duration(	CelluloidControlBox *box,
 						gint duration )
 {
 	celluloid_seek_bar_set_duration(CELLULOID_SEEK_BAR(box->seek_bar), duration);
+	celluloid_seek_bar_set_duration(CELLULOID_SEEK_BAR(box->secondary_seek_bar), duration);
 }
 
 void
@@ -723,28 +927,33 @@ celluloid_control_box_get_volume_popup_visible(CelluloidControlBox *box)
 }
 
 void
-celluloid_control_box_set_fullscreen_state(	CelluloidControlBox *box,
-						gboolean fullscreen )
+celluloid_control_box_set_floating(CelluloidControlBox *box, gboolean floating)
 {
-	GtkWidget *image =
-		gtk_button_get_image(GTK_BUTTON(box->fullscreen_button));
+	if(floating)
+	{
+		gtk_style_context_add_class
+			(gtk_widget_get_style_context(GTK_WIDGET(box)), "osd");
 
-	gtk_image_set_from_icon_name(	GTK_IMAGE(image),
-					fullscreen?
-					"view-restore-symbolic":
-					"view-fullscreen-symbolic",
-					GTK_ICON_SIZE_BUTTON );
+		gtk_widget_set_margin_start(GTK_WIDGET(box), 12);
+		gtk_widget_set_margin_end(GTK_WIDGET(box), 12);
+		gtk_widget_set_margin_top(GTK_WIDGET(box), 12);
+		gtk_widget_set_margin_bottom(GTK_WIDGET(box), 12);
+	}
+	else
+	{
+		gtk_style_context_remove_class
+			(gtk_widget_get_style_context(GTK_WIDGET(box)), "osd");
 
-	gtk_widget_set_visible(	box->fullscreen_button,
-				box->show_fullscreen_button );
+		gtk_widget_set_margin_start(GTK_WIDGET(box), 0);
+		gtk_widget_set_margin_end(GTK_WIDGET(box), 0);
+		gtk_widget_set_margin_top(GTK_WIDGET(box), 0);
+		gtk_widget_set_margin_bottom(GTK_WIDGET(box), 0);
+	}
 }
 
 void
 celluloid_control_box_reset(CelluloidControlBox *box)
 {
 	celluloid_control_box_set_seek_bar_pos(box, 0);
-	celluloid_control_box_set_seek_bar_duration(box, 0);
-	set_playing_state(box, FALSE);
-	set_skip_enabled(box, FALSE);
-	celluloid_control_box_set_fullscreen_state(box, FALSE);
+	set_fullscreen_state(box, FALSE);
 }
